#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Nov 2018
#
# _________________________________________________________________________
# various tools machine learning data handling
#

import os
import numpy as np

import odpy.dbman as oddbman
import dgbpy.keystr as dgbkeys
import dgbpy.hdf5 as dgbhdf5

nladbdirid = '100060'
mlinpgrp = 'Deep Learning Example Data'
mltrlgrp = 'Deep Learning Model'
dgbtrl = 'dGB'

def getInfo( filenm, quick=False ):
  return dgbhdf5.getInfo( filenm, quick )

def datasetCount( dsets ):
  if dgbkeys.datasetdictstr in dsets:
    dsets = infos[dgbkeys.datasetdictstr]
  ret = {}
  totsz = 0
  for groupnm in dsets:
    collection = dsets[groupnm]
    collcounts = {}
    groupsz = 0
    for collnm in collection:
      collitm = collection[collnm]
      collsz = len(collitm)
      groupsz += collsz
      collcounts.update({ collnm: collsz })
    collcounts.update({ 'size': groupsz })
    totsz += groupsz
    ret.update({groupnm: collcounts})
  ret.update({ 'size': totsz })
  return ret

def getDatasetNms( dsets, validation_split=None, valid_inputs=None ):
  train = {}
  valid = {}
  if validation_split == None:
    validation_split = 0
  dorandom = True
  if validation_split > 1:
    validation_split = 1
  elif validation_split < 0:
    validation_split = 0
  elif validation_split == 0:
    dorandom = False
  if valid_inputs == None:
    valid_inputs = {}
  for groupnm in dsets:
    group = dsets[groupnm]
    traingrp = {}
    validgrp = {}
    for inp in group:
      dsetnms = group[inp].copy()
      nrpts = len(dsetnms)
      if dorandom:
        nrpts = int(nrpts*validation_split)
        np.random.shuffle( dsetnms )
      if inp in valid_inputs:
        if dorandom:
          traingrp.update({inp: dsetnms[nrpts:]})
          validgrp.update({inp: dsetnms[:nrpts]})
        else:
          validgrp.update({inp: dsetnms})
      else:
        if dorandom and len(valid_inputs)<1:
          traingrp.update({inp: dsetnms[nrpts:]})
          validgrp.update({inp: dsetnms[:nrpts]})
        else:
          traingrp.update({inp: dsetnms})
    train.update({groupnm: traingrp})
    valid.update({groupnm: validgrp})
  return {
    dgbkeys.traindictstr: train,
    dgbkeys.validdictstr: valid
  }

def getChunks(dsets,nbchunks):
  ret = []
  for ichunk in range(nbchunks):
    datagrp = {}
    for groupnm in dsets:
      alldata = dsets[groupnm]
      surveys = {}
      for inp in alldata:
        datalist = alldata[inp]
        nrpts = len(datalist)
        start = int(ichunk * nrpts / nbchunks)
        end = int((ichunk+1) * nrpts / nbchunks)
        datalist = datalist[start:end]
        surveys.update({inp: datalist})
      datagrp.update({groupnm: surveys})
    ret.append(datagrp)
  return ret
  
def hasScaler( infos, inputsel=None ):
  inp = infos[dgbkeys.inputdictstr]
  for inputnm in inp:
    if inputsel != None and not inputnm in inputsel:
      continue
    if not dgbkeys.scaledictstr in inp[inputnm]:
      return False
  return True

def getDatasetsByGroup( dslist, groupnm ):
  ret = {}
  for keynm in dslist:
    dp = dslist[keynm]
    if groupnm in dp:
      ret.update({keynm:{groupnm: dp[groupnm]}})
  return ret

def getSomeDatasets( dslist, decim=None ):
  if decim == None or decim <= 0 or decim==False:
    return dslist
  if decim > 1:
    decim = 1
  ret = {}
  for dsetnm in dslist:
    sret = {}
    dset = dslist[dsetnm]
    for groupnm in dset:
      group = dset[groupnm]
      setgrp = {}
      for inp in group:
        dsetnms = group[inp].copy()
        nrpts = int(len(dsetnms)*decim)
        np.random.shuffle( dsetnms )
        del dsetnms[nrpts:]
        setgrp[inp] = dsetnms
      sret[groupnm] = setgrp
    ret[dsetnm] = sret
  return ret

def getTrainingData( filenm, decim=False ):
  infos = getInfo( filenm )
  return getTrainingDataByInfo( infos )

def getTrainingDataByInfo( info, dsetsel=None ):
  examples = dgbhdf5.getDatasets( info, dsetsel )
  ret = {}
  for ex in examples:
    ret.update({ex: examples[ex]})
  y_examples = list()
  if dgbkeys.ytraindictstr in examples:
    y_examples.append( examples[dgbkeys.ytraindictstr] )
  if dgbkeys.yvaliddictstr in examples:
    y_examples.append( examples[dgbkeys.yvaliddictstr] )
  ret.update({ dgbkeys.infodictstr: getClasses(info,y_examples) })
  if dgbkeys.classdictstr in info and info[dgbkeys.classdictstr]:
    if dgbkeys.ytraindictstr in examples:
      normalize_class_vector( examples[dgbkeys.ytraindictstr], \
                              info[dgbkeys.classesdictstr] )
    if dgbkeys.yvaliddictstr in examples:
      normalize_class_vector( examples[dgbkeys.yvaliddictstr], \
                              info[dgbkeys.classesdictstr] )
  return ret

def getClasses( info, y_vectors ):
  if not info[dgbkeys.classdictstr] or dgbkeys.classesdictstr in info:
    return info
  import numpy as np
  classes = []
  for y_vec in y_vectors:
    (minval,maxval) = ( np.min(y_vec), np.max(y_vec) )
    for idx in np.arange(minval,maxval+1,1,dtype=np.uint8):
      if np.any(y_vec == idx ):
        classes.append( idx )
  if len(classes) > 0:
    info.update( {dgbkeys.classesdictstr: np.array(classes,dtype=np.uint8)} )
  return info

def normalize_class_vector( arr, classes ):
  import numpy as np
  classes = np.sort( classes )
  for i in range( len(classes) ):
    arr[arr == classes[i]] = i

def unnormalize_class_vector( arr, classes ):
  import numpy as np
  classes = np.sort( classes )
  for i in reversed(range( len(classes) ) ):
    arr[arr == i] = classes[i]

def saveModel( model, inpfnm, platform, infos, outfnm ):
  from odpy.common import log_msg
  if os.path.exists(outfnm):
    try:
      os.remove( outfnm )
    except Exception as e:
      log_msg( '[Warning] Could not remove pre-existing model file:', e )
  log_msg( 'Saving model.' )
  if platform == dgbkeys.kerasplfnm:
    import dgbpy.dgbkeras as dgbkeras
    dgbkeras.save( model, outfnm )
  elif platform == dgbkeys.scikitplfnm:
    import dgbpy.dgbscikit as dgbscikit
    dgbscikit.save( model, outfnm )
  else:
    log_msg( 'Unsupported machine learning platform' )
    raise AttributeError
  dgbhdf5.addInfo( inpfnm, platform, outfnm, infos, model.__class__.__name__ )
  log_msg( 'Model saved.' )

def getModel( modelfnm, fortrain=False ):
  infos = getInfo( modelfnm )
  platform = infos[dgbkeys.plfdictstr]
  if platform == dgbkeys.kerasplfnm:
    import dgbpy.dgbkeras as dgbkeras
    model = dgbkeras.load( modelfnm, fortrain )
  elif platform == dgbkeys.scikitplfnm:
    import dgbpy.dgbscikit as dgbscikit
    model = dgbscikit.load( modelfnm )
  else:
    from odpy.common import log_msg
    log_msg( 'Unsupported machine learning platform' )
    raise AttributeError
  return (model,infos)

def getApplyInfoFromFile( modelfnm, outsubsel=None ):
  return getApplyInfo( getInfo(modelfnm), outsubsel )

def getApplyInfo( infos, outsubsel=None ):
  isclassification = infos[dgbkeys.classdictstr]
  firstoutnm = dgbhdf5.getMainOutputs(infos)[0]
  if isclassification:
    names = firstoutnm
    preddtype = 'uint8'
  else:
    names = []
    preddtype = 'float32'

  probdtype = 'float32'
  confdtype = 'float32'
  if outsubsel != None:
    if 'targetnames' in outsubsel:
      names = outsubsel['targetnames']
    if dgbkeys.dtypepred in outsubsel:
      preddtype = outsubsel[dgbkeys.dtypepred]
    if dgbkeys.dtypeprob in outsubsel:
      probdtype = outsubsel[dgbkeys.dtypeprob]
    if dgbkeys.dtypeconf in outsubsel:
      confdtype = outsubsel[dgbkeys.dtypeconf]

  withpred = (isclassification and firstoutnm in names) or \
             not isclassification
  if isclassification:
    (withprobs,classnms) = dgbhdf5.getClassIndices( infos, names )
  else:
    withprobs = []
  ret = {
    dgbkeys.classdictstr: isclassification,
  }
  withconfidence = isclassification and dgbkeys.confvalstr in names

  if withpred:
    ret.update({dgbkeys.dtypepred: preddtype})
  if isclassification:
    if len(withprobs) > 0:
      ret.update({
        dgbkeys.probadictstr: withprobs,
        dgbkeys.dtypeprob: probdtype
      })
    if withconfidence:
      ret.update({dgbkeys.dtypeconf: confdtype})

  return ret

dblistall = None

def modelNameIsFree( modnm, type, args, reload=True ):
  (exists,sametrl,sameformat,sametyp) = \
             modelNameExists( modnm, type, args, reload )
  if exists == False:
    return True
  if not sametrl or not sameformat:
    return False
  if sametyp != None:
    return sametyp
  return False

def modelNameExists( modnm, type, args, reload=True ):
  (sametrl,sameformat,sametyp) = (None,None,None)
  modinfo = dbInfoForModel( modnm, args, reload )
  exists = modinfo != None
  if exists:
    sameformat = modinfo['Format'] == dgbtrl
    if 'Type' in modinfo:
      sametyp = type == modinfo['Type']
    if 'TranslatorGroup' in modinfo:
      sametrl = modinfo['TranslatorGroup'] == mltrlgrp
  return (exists,sametrl,sameformat,sametyp)

def dbInfoForModel( modnm, args, reload=True ):
  global dblistall
  if dblistall == None or reload:
    dblistall = oddbman.getDBList(mltrlgrp, alltrlsgrps=True, args=args)
  return oddbman.getInfoFromDBListByNameOrKey( modnm, dblistall )

def getModelType( infos ):
  return infos[dgbkeys.learntypedictstr]

def getSaveLoc( outnm, ftype, args ):
  dblist = oddbman.getDBList(mltrlgrp,alltrlsgrps=False, args=args)
  try:
    dbkey = oddbman.getDBKeyForName( dblist, outnm )
  except ValueError:
    return oddbman.getNewEntryFileName(outnm,nladbdirid,mltrlgrp,dgbtrl,\
                                       dgbhdf5.hdf5ext,ftype=ftype,args=args)
  return oddbman.getFileLocation( dbkey, args )



"""Common tools for odpy package

Copyright (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
  * AUTHOR : A. Huck
  * DATE   : July 2018

"""

import sys
import os
import platform
import logging
import tempfile
from datetime import datetime
import threading

def sTimeUnitString( ismilli=False, abbr=True ):
  """OpendTect-like time stamp

  Parameters:
    * ismilli (bool, optional): Include millisecond (default is False)
    * abbr (bool, optional): Abbreviated (default is True)

  Returns:
    * str: Time stamp string formatted like done by OpendTect

  Examples:
    >>> sTimeUnitString()
    'Mon 20 Apr 2020, 13:59:54'

    >>> sTimeUnitString( True )
    'Mon 20 Apr 2020, 13:59:54.001245'

    >>> sTimeUnitString( True, True )
    'Mon 20 Apr 2020, 13:59:54.001245'

    >>> sTimeUnitString( True, False )
    'Monday 20 April 2020, 13:59:54'

  """

  if abbr:
    fmt = "%a %d %b"
  else:
    fmt = "%A %d %B"
  fmt += " %Y, %X"
  if ismilli:
    fmt += ".%f"
  return datetime.now().strftime(fmt)

try:
  from bokeh.util import logconfig
except ImportError:
  logconfig = None

syslog_logger = logging.getLogger(__name__)
proclog_logger = logging.getLogger('odproclog')

if not syslog_logger.hasHandlers():
  handler = logging.StreamHandler(sys.stdout)
  syslog_logger.setLevel( 'INFO' )
  syslog_logger.addHandler( handler )

if not proclog_logger.hasHandlers():
  handler = logging.StreamHandler(sys.stdout)
  proclog_logger.setLevel( 'DEBUG' )
  proclog_logger.addHandler( handler )

def initLogging(args):
  """odpy Logger initialization

  Parameters:
    * args (dict):
      The members 'logfile' and 'sysout' from the input dictionary are supposed to contain existing filenames which are used
      to setup the module loggers proclog_logger and syslog_logger

  """

  set_log_file( args['logfile'].name, proclog_logger )
  set_log_file( args['sysout'].name, syslog_logger )

def set_log_file( filenm, logger ):
  for handler in logger.handlers:
    logger.removeHandler( handler )
  if filenm == '<stdout>':
    handler = logging.StreamHandler(sys.stdout)
    logger.addHandler( handler )
    return
  elif filenm == '<stderr>':
    handler = logging.StreamHandler(sys.stderr)
    logger.addHandler( handler )
    return
  if not os.path.exists(filenm):
    std_msg( 'Log file not found: ', filenm )
    return
  handler = logging.FileHandler(filenm,'a')
  logger.addHandler( handler )
  logger.propagate = False

def get_log_logger():
  return proclog_logger

def get_std_logger():
  return syslog_logger

def mergeArgs(a,b=None,c=None,d=None,e=None,f=None):
  msg = str(a)
  if b != None:
    msg = msg+' '+str(b)
  if c != None:
    msg = msg+' '+str(c)
  if d != None:
    msg = msg+' '+str(d)
  if e != None:
    msg = msg+' '+str(e)
  if f != None:
    msg = msg+' '+str(f)
  return msg

def std_msg(a,b=None,c=None,d=None,e=None,f=None):
  """Print to odpy standard logger

  Parameters:
    * a (object or string): Message to be printed
    * b-f (object or string, optional): Message to be printed

  Returns:
    * str: Concatenated string

  Notes:
    * All objects are formatted to strings using the str() function
    * All outputs are automatically separated by spaces.
    * Reserved for standard logging information.

  """

  msg = mergeArgs(a,b,c,d,e,f)
  get_std_logger().info(msg)

def log_msg(a,b=None,c=None,d=None,e=None,f=None):
  """Print to odpy processing logger

  Parameters:
    * a (object or string): Message to be printed
    * b-f (object or string, optional): Message to be printed

  Returns:
    * str: Concatenated string

  Notes:
    * All objects are formatted to strings using the str() function
    * All outputs are automatically separated by spaces.
    * Reserved for processing logging information,
    * i.e. to report progress on a task

  """
  msg = mergeArgs(a,b,c,d,e,f)
  get_log_logger().debug(msg)

def has_file_handlers(logger):
  for handler in logger.handlers:
    if isinstance( handler, logging.FileHandler ):
      return True
  return False

def has_stdlog_file():
  return has_file_handlers( get_std_logger() )

def has_log_file():
  return has_file_handlers( get_log_logger() )

def get_handler_stream(logger):
  for handler in logger.handlers:
    if isinstance( handler, logging.StreamHandler ) and handler.level != logging.NOTSET:
      return handler.stream
    elif isinstance( handler, logging.FileHandler ) and handler.level != logging.NOTSET:
      return handler.stream
  return None

def get_handler_filename(logger):
  for handler in logger.handlers:
    if isinstance( handler, logging.FileHandler ):
      return handler.baseFilename
  return None

def get_std_stream():
  return get_handler_stream( get_std_logger() )

def get_log_stream():
  return get_handler_stream( get_log_logger() )

def get_stdlog_file():
  return get_handler_filename( get_std_logger() )

def get_log_file():
  return get_handler_filename( get_log_logger() )

def reset_log_file( keeplines=0 ):
  """Log file reset

  Parameters:
    * keeplines (int, optional): 
      Number of lines from the top of the file to keep (default is 0)

  Empty the log file pointed at by the processing logger,
  for instance before starting a new task.

  """

  if not has_log_file():
    return
  logfnm = get_log_file()
  idx = 0
  f = open( logfnm, 'r' )
  keptlines = list()
  for line in f:
    keptlines.append( line )
    idx = idx + 1
    if idx >= keeplines:
      break
  f.close()
  f = open( logfnm, 'w' )
  for line in keptlines:
    f.write( line )
  f.close()
  set_log_file( logfnm, proclog_logger )

def redirect_stdout():
  """Stdout-stderr redirection

  * Forces stdout to point to odpy.proclog_logger
  * Forces stderr to point to odpy.syslog_logger

  Notes:
    * Changes the value of sys.stdout and sys.stderr
    * Should be avoided as much as possible: To be used only when one cannot
      redirect in any other way.

  """
  if (logconfig is  None) or (not logging.getLogger() == logconfig.root_logger):
    return
  if has_log_file():
    sys.stdout = open( get_log_file(), 'a' )
  if has_stdlog_file():
    sys.stderr = open( get_stdlog_file(), 'a' )

def restore_stdout():
  """Stdout-stderr restore

  Undo operation of redirect_stdout

  """

  if (logconfig is None) or (not logging.getLogger() == logconfig.root_logger):
    return
  if has_log_file():
    sys.stdout = sys.__stdout__
  if has_stdlog_file():
    sys.stderr = sys.__stderr__

def isWin():
  """Is platform Windows?

  Returns:
    * True if running on any Windows platform

  """

  return platform.system() == 'Windows'

def isMac():
  """Is platform Mac?

  Returns:
    * True if running on any Mac-OS platform

  """

  return platform.system() == 'Darwin'

if platform.python_version() < "3":
  std_msg( "odpy requires at least Python 3" )
  sys.exit( 1 )

def getPlfSubDir():
  """Platform sub-directory

  Platform specific sub-directory as existing
  in an OpendTect software installation
  
  Returns:
    * str: string like: 'lux64', 'win64', 'mac'.
      None if the platform is not supported by OpendTect

  """

  system = platform.system()
  arch = platform.architecture()[0]
  if system == 'Linux':
    if arch == '64bit':
      return 'lux64'
    else:
      return None
  elif system == 'Windows':
    if arch == '64bit':
      return 'win64'
    else:
      return 'win32'
  elif system == 'Darwin':
    if arch == '64bit':
      return 'mac'
    else:
      return None
  else:
    return None

def getBinSubDir():
  """Binary sub-directory

  Sub-directory containing the executables in an OpendTect installation

  Returns:
    * str: string like: 'Debug', 'Release', 'RelWithDebInfo'. 
      None if no such sub-directory is found.

  Notes:
    Assumes that the OpendTect installation contains the 
    executable 'od_FileBrowser', and that this installation can
    be located by the function getODSoftwareDir()

  """

  if isMac():
    return None
  plfsubdirfp = os.path.join( getODSoftwareDir(), 'bin', getPlfSubDir() )
  execnm = 'od_FileBrowser'
  if isWin():
    execnm = execnm+'.exe'
  reltypes = ('Debug','Release','RelWithDebInfo')
  for reltype in reltypes:
    if os.path.isfile( os.path.join(plfsubdirfp,reltype,execnm) ):
      return reltype
  return None

def getODSoftwareDir(args=None):
  """OpendTect sofware directory

  Parameters:
    * args (dict, optional):
      Dictionary with the member 'dtectexec'. The value
      for that member should point to the executables folder
      of the requested application

  Returns:
    * str: Full path to the OpendTect software installation

  Notes:
    * Retrieved from either the input dictionary
      or from the current environment by reading
      the 'DTECT_APPL' or 'DTECT_WINAPPL' variables which
      are set by OpendTect at runtime.
    * Neither dictionary nor environment variables need to be set
      if the current module is placed within an OpendTect installation

  Examples:
    >>> getODSoftwareDir()
    'C:\\Program Files\\OpendTect\\6.6.0'

  """

  if args != None and 'dtectexec' in args:
    bindir = getExecPlfDir(args)
    return bindir
  applenvvar = 'DTECT_APPL'
  if isWin():
    applenvvar = 'DTECT_WINAPPL'
  if applenvvar in os.environ:
    return os.environ[applenvvar]
  curdir = os.path.dirname( __file__ )
  maxrecur = 15
  relinfodir = 'relinfo'
  if isMac():
    relinfodir = os.path.join('Resources',relinfodir)
  while not os.path.isdir(os.path.join(curdir,relinfodir)) and maxrecur > 0:
    curdir = os.path.dirname( curdir )
    maxrecur = maxrecur-1
  return curdir

def getExecPlfDir(args=None):
  """OpendTect executables directory

  Parameters:
    * args (dict, optional):
      Dictionary with the member 'dtectexec'. The value
      for that member should point to the executables folder
      of the requested application

  Returns:
    * str: Full path to the binaries of an OpendTect installation

  Notes:
    * Assumes that the software installation can be located with getODSoftwareDir()
      and that python is running on a platform supported by OpendTect.

  Examples:
    >>> getExecPlfDir()
    'C:\\Program Files\\OpendTect\\6.6.0\\bin\\win64\\Release'

  """

  if args != None and 'dtectexec' in args:
    return args['dtectexec'][0]
  appldir = getODSoftwareDir()
  if isMac():
    return os.path.join( getODSoftwareDir(), 'Contents', 'MacOS' )
  else:
    return os.path.join( getODSoftwareDir(), 'bin', getPlfSubDir(), getBinSubDir())

def getODArgs(args=None):
  """OpendTect arguments dictionary

  Create a dictionary that contains typical OpendTect
  command line arguments and the files from this module' loggers

  Parameters:
    * args (dict, optional):
      Dictionary with the member 'dtectexec'. The value
      for that member should point to the executables folder
      of the requested application

  Returns:
    * dict: A dictionary with the following key-values:
        * 'dtectexec' : Full path to the OpendTect installation (see getExecPlfDir)
        * 'survey' : The survey directory name, if provided by the input dictionary
        * 'proclog' : The log file from proclog_logger if applicable
        * 'syslog' : The log file frol syslog_logger if applicable

  """

  ret = {
    'dtectexec': [getExecPlfDir(args)]
  }
  if args != None and 'dtectdata' in args:
    ret.update({'dtectdata': args['dtectdata']})
  if args != None and 'survey' in args:
    ret.update({'survey': args['survey']})
  if has_log_file() :
    ret.update({'proclog': args['logfile'].name})
  if has_stdlog_file():
    ret.update({'syslog': args['sysout'].name})
#  else:
#    if has_log_file():
#      ret.update({'syslog': args['logfile'].name})
  return ret

def getIconFp(nm,args=None):
  """Path to an OpendTect icon file

  Parameters:
    * nm (string): Icon file basename (without extension)
    * args (dict, optional)
      Dictionary with the member 'dtectexec'. The value
      for that member should point to the executables folder
      of the requested application

  Returns:
    * str: Full path to the icon folder of the OpendTect installation

  Notes:
    Assumes that the software installation can be located with getODSoftwareDir()

  Examples:
    >>> getIconFp( 'dgbpro' )
    'C:\\Program Files\\OpendTect\\6.6.0\\data\\icons.Default\\dgbpro.png'
    
  """
  oddir = getODSoftwareDir(args)
  ret = os.path.join(oddir,'data','icons.Default',nm)+'.png'
  if os.path.exists(ret):
    return ret
  return None

def tail(fp,lines=1,strip_empty=False,_buffer=4098):
  lines_found = []
  block_counter = -1
  while len(lines_found) < lines:
    try:
      fp.seek(block_counter * _buffer, os.SEEK_END)
    except IOError:
      fp.seek(0)
      lines_found = fp.readlines()
      break

    lines_found = fp.readlines()
    block_counter -= 1

  ret = lines_found[-lines:]
  if strip_empty:
    while '\n' in ret:
      ret.remove('\n')
  return ret

def batchIsFinished( logfile ):
  """OpendTect batch processing status

  Checks if an OpendTect batch process reached completion by
  parsing its log file.

  Parameters:
    logfile (str): Full path to an existing OpendTect log file.

  Returns:
    * bool:
      True if the file exists and contains 'Finished batch processing' at its tail.

  Notes:
    Instantaneous status. To monitor if the processing finishes,
    the user must repeatedly call this function.

  """

  ret = list()
  with open(logfile) as fd:
    ret = tail(fd,10,True)
  return len(ret) > 0 and 'Finished batch processing' in ret[-1]

def writeFile( fnm, content ):
  try:
    f = open( fnm, 'w' )
    f.write( content )
    f.close()
  except:
    return False
  return True

def getTempDir():
    # This HAS to correspond with the C++ getTempDir() function
    return tempfile.gettempdir()

class Timer(threading.Timer):
    """Repeated timer

    Timer that restarts automatically after each interval
    """

    def run(self):
        while not self.finished.is_set():
            self.finished.wait(self.interval)
            self.function(*self.args, **self.kwargs)
        self.finished.set()



import numpy as np
import json

import odpy.common as odcommon
import odpy.dbman as oddbman
from odpy.oscommand import getODCommand, execCommand

wellmanexe = 'od_WellMan'
wlldbdirid = '100050'
wlltrlgrp = 'Well'
dgbtrl = 'dGB'

dblist = None

def getNames( reload=False, args=None ):
  if args == None:
    args = odcommon.getODArgs()
  return getWellDBList(reload,args)['Names']

def getInfo( wllnm, reload=False, args=None ):
  dbkey = getDBKey( wllnm, reload=reload, args=args )
  return oddbman.getInfoByName( dbkey, wlltrlgrp,exenm=wellmanexe, args=args )

def getName( dbkey, reload=False, args=None ):
  cmd = getODCommand(wellmanexe,args)
  cmd.append( '--info' )
  cmd.append( dbkey )
  ret = oddbman.getDBDict( cmd )
  return ret['Name']

def getLogNames( wllnm, reload=False, args=None ):
  dbkey = getDBKey( wllnm, reload=reload, args=args )
  cmd = getODCommand(wellmanexe,args)
  cmd.append( '--list-logs' )
  cmd.append( dbkey )
  ret = oddbman.getDBDict( cmd )
  return ret['Names']

def getLog( wllnm, lognm, reload=False, args=None ):
  dbkey = getDBKey( wllnm, reload=reload, args=args )
  cmd = getODCommand(wellmanexe,args)
  cmd.append( '--read-log' )
  cmd.append( dbkey )
  cmd.append( lognm )
  ret = oddbman.getDBDict( cmd )
  return (ret['MDs'], ret['Values'])

def getDBKey( wllnm, reload=False, args=None ):
  global dblist
  dblist = getWellDBList(reload,args)
  return oddbman.getDBKeyForName( dblist, wllnm )  

def getWellDBList( reload, args=None ):
  global dblist
  if dblist != None and not reload:
    return dblist

  if args == None:
    args = odcommon.getODArgs()
  dblist = oddbman.getDBList(wlltrlgrp,exenm=wellmanexe,args=args)
  return dblist

def getMarkers( wllnm, reload=False, args=None ):
  dbkey = getDBKey( wllnm, reload=reload, args=args )
  cmd = getODCommand(wellmanexe,args)
  cmd.append( '--list-markers' )
  cmd.append( dbkey )
  ret = oddbman.getDBDict( cmd )
  return (ret['Names'], ret['MDs'], ret['Color'])

def getTrack( wllnm, reload=False, args=None ):
  dbkey = getDBKey( wllnm, reload=reload, args=args )
  cmd = getODCommand(wellmanexe, args)
  cmd.append( '--read-track' )
  cmd.append( dbkey )
  ret = oddbman.getDBDict( cmd )
  return (ret['MDs'], ret['TVDs'], ret['X-Coords'], ret['Y-Coords'])




from numba import cuda as numcuda

def is_cuda_available():
  return numcuda.is_available()

def detect_gpus(printfn=print):
  printfn( numcuda.detect() )

def nr_gpus():
  return len(numcuda.gpus)

def compute_capability( devnum=0 ):
  return numcuda.gpus[devnum].compute_capability

def id( devnum=0 ):
  return numcuda.gpus[devnum].id

def name( devnum=0 ):
  return numcuda.gpus[devnum].name

def get_memory_info( devnum=None ):
  return numcuda.current_context(devnum=devnum).get_memory_info()



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : Wayne Mogg
# DATE     : January 2020
#
# Service Manager
#
# 
import json
import os
import io
import psutil
import signal
import socket
import struct
import sys
import threading
import odpy.common as odcommon
from tornado.iostream import StreamClosedError
from tornado import gen
import tornado.tcpserver

class ServiceMgr(tornado.tcpserver.TCPServer):
  def __init__(self, cmdserver, ppid, tornadoport, serviceID=None):
    super(ServiceMgr, self).__init__()
    self.cmdserver = cmdserver
    self.cmdhost = None
    self.cmdport = None
    self.serviceID = serviceID
    if '@' in cmdserver:
      info = cmdserver.split('@')
      cmdserver = info[1]
    if ':' in cmdserver:
      info = cmdserver.split(':')
      self.cmdhost = info[0]
      self.cmdport = int(info[1])

    self._parentproc = None
    if ppid > 0:
      self._parentproc = psutil.Process(ppid)
      tornado.ioloop.PeriodicCallback(self._parentChkCB, 1000)

    self._startServer( tornadoport )
    self._actions = dict()
    
  def __enter__(self):
    return self
  
  def __exit__(self, exc_type, exc_value, traceback):
    pass
#    self.stop()

  def _is_port_in_use(self, port, local_ip):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
      if s.connect_ex(('127.0.0.1', port)) == 0:
        return True
      retval = False
      try:
        retval = s.connect_ex((local_ip, port)) == 0
      except:
        pass
      return retval
    
  def _startServer(self, tornadoport, attempts=20):
    address = 'localhost'
    port = max(self.cmdport+1,tornadoport+1)
    hostname = socket.gethostname()
    local_ip = socket.gethostbyname( hostname )
    while attempts:
      attempts -=1
      if self._is_port_in_use(port,local_ip):
        odcommon.std_msg( 'Port in use:', port )
        port += 1
        continue
      try:
        self.listen(port,address=address)
        self._register(port,address)
        return
      except OSError as ex:
# using error code since the error string is translated in current locale
        if '10048' in str(ex):
          odcommon.std_msg( 'Failed to listen to port:', port )
          port += 1
        else:
          raise ex
    raise Exception("Failed to find available port")

  def _register(self, port, address):
#    odcommon.std_msg( 'Registering with port', port, 'and address', address )
    Message().sendObject(self.cmdhost, self.cmdport,
                   'bokeh_register', {'servicename': self.serviceID,
                                'hostname': address,
                                'port': port,
                                'pid': os.getpid()
                                })
    
  def _parentChkCB(self):
    if self._parentproc != None and not self._parentproc.is_running():
      odcommon.std_msg('Found dead parent, exiting')
      self.stop()
      os.kill(psutil.Process().pid, signal.SIGINT)
      
  async def handle_stream(self, stream, address):
    hdrlen = 10
    while True:
      try:
        odhdr = await stream.read_bytes(hdrlen)
        payload_len = struct.unpack('=i', odhdr[0:4])[0]
        packetbody = await stream.read_bytes(payload_len)
        inpacket = Packet(odhdr + packetbody)
        resp_packet = self._processPacket(inpacket)
        await stream.write(resp_packet.packet)
        stream.close()
      except StreamClosedError:
        break
      
  def _processPacket(self, inpacket):
    payload = inpacket.getTextPayload()
    result = {}
    for key, params in payload.items():
      result = self._actions.get(key)(params)
      
    obj = dict()
    obj[key] = result
    inpacket.setTextPayload(obj)
    return inpacket

  def addAction(self, key, action):
    self._actions[key] = action

  def sendObject(self, objkey, jsonobj):
      msgobj = {'bokehid': self.serviceID}
      msgobj.update(jsonobj)
      Message().sendObjectToAddress(self.cmdserver, objkey, msgobj)

    
class Message:
  def parseAddress(self, address):
    host = None
    port = None
    if '@' in address:
      info = address.split('@')
      address = info[1]
    if ':' in address:
      info = address.split(':')
      host = info[0]
      port = int(info[1])
    return host, port
        
  def sendObject(self, host, port, objkey, jsonobj):
    packet = Packet()
    packet.setIsNewRequest()
    obj = dict()
    obj[objkey] = jsonobj
    packet.setTextPayload(obj)
    tornado.ioloop.IOLoop.current().add_callback(self._send, host, port, packet)
    
  def sendObjectToAddress(self, address, objkey, jsonobj):
    host, port = self.parseAddress(address)
    self.sendObject(host, port, objkey, jsonobj)

  def sendEvent(self, host, port, eventstr):
    packet = Packet()
    packet.setIsNewRequest()
    action = {'action': eventstr}
    packet.setTextPayload(action)
    tornado.ioloop.IOLoop.current().add_callback(self._send, host, port, packet)

  def sendEventToAddress(self, address, eventstr):
    host, port = self.parseAddress(address)
    self.sendEvent(host, port, eventstr)

  async def _send(self, host, port, packet):
    client = tornado.tcpclient.TCPClient()
    stream = await client.connect(host, port)
    await stream.write(packet.packet)
    stream.close()


class Packet:
  _curreqid = 0
  def __init__(self, packet=None):
    self._reqid = None
    self._subid = None
    self.jsonheader = None
    self.packet = packet
    self._lock = threading.Lock()

  def setIsNewRequest(self):
    with self._lock:
      self._curreqid += 1
    self._reqid = self._curreqid
    self._subid = -1
    return self._reqid
    
  def setTextPayload(self, jsonobj):
    content_encoding = 'utf-8'
    payload = {
          'content_bytes': self._json_encode(jsonobj, content_encoding),
          'content_type': 'text/json',
          'content_encoding': content_encoding,
          'arrsize': None,
    }
    self._createPacket(payload)
    
  def getTextPayload(self):
    content_encoding = 'utf-8'
    payload_bytes = self._odhdr_decode()
    self.jsonheader, data = self._json_decode(payload_bytes, content_encoding)

    for reqhdr in (
        "byteorder",
        "content-length",
        "content-type",
        "content-encoding",
      ):
      if reqhdr not in self.jsonheader:
        raise ValueError(f'Missing required header "{reqhdr}".')
    
    content_len = self.jsonheader['content-length']
    if len(data)!=content_len:
        raise ValueError(f'Message payload size error, expected "{content_len}" got "{len(data)}".')
    
    if self.jsonheader['content-type']=='text/json':
      payload, data = self._json_decode(data, content_encoding)
      
    return payload

  def _createPacket(self, payload):
    jsonheader = {
        'byteorder': sys.byteorder,
        'content-type': payload['content_type'],
        'content-encoding': payload['content_encoding'],
        'content-length': len(payload['content_bytes']),
        }
    if payload['arrsize'] != None:
      jsonheader.update({ 'array-shape': payload['arrsize'] })

    jsonheader_bytes = self._json_encode(jsonheader, 'utf-8')
    payload_bytes = jsonheader_bytes + payload['content_bytes']
    od_hdr = struct.pack('=i',len(payload_bytes)) \
             + struct.pack('=i',self._reqid) \
             + struct.pack('=h',self._subid)
    self.packet = od_hdr + payload_bytes

  def _json_encode(self, obj, encoding):
    json_hdr = json.dumps(obj, ensure_ascii=False).encode(encoding)
    return struct.pack('=i',len(json_hdr)) + json_hdr

  def _odhdr_decode(self):
    hdrlen = 10
    if self.packet==None or len(self.packet)<hdrlen:
      return False
    self._reqid = struct.unpack('=i',self.packet[4:8])[0]
    self._subid = struct.unpack('=h',self.packet[8:hdrlen])[0]
    return self.packet[hdrlen:]
    
    
  def _json_decode(self, json_bytes, encoding):
    hdrlen = 4
    jsonobj_len = struct.unpack('=i',json_bytes[:hdrlen])[0]
    jsonbytes = json_bytes[hdrlen:hdrlen+jsonobj_len]
    jsonstr = jsonbytes.decode('utf-8')
    if odcommon.isWin():
      jsonstr = jsonstr.translate(str.maketrans({"\\": r"\\"}))
    jsonobj = json.loads( jsonstr )

    return (jsonobj, json_bytes[hdrlen+jsonobj_len:])



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dgbpy-dm',
    version='0.1.2',
    author='David Markus',
    author_email='david.datascientist@outlook.com',
    description='Useful dgb stuff.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='http://opendtect.com',
    packages=setuptools.find_packages(),
    install_requires=[
        "bokeh>=2.1.0",
        "psutil>=5.8.0",
        "h5py>=3.1.0",
        "joblib>=1.0.0",
        "sklearn>=0.0",
        "jedi>=0.16",
        "tensorflow==2.1.0",
        "keras",
        "humanfriendly>=9.1",
        "xgboost>=1.3.3",
        "flake8",
        "autopep8",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : Wayne Mogg
# DATE     : January 2020
#
# Support methods for embedded Bokeh Server
#
# 
import os
from bokeh.util.logconfig import basicConfig
from bokeh.server.server import Server
import dgbpy.servicemgr as dgbservmgr

def DefineBokehArguments(parser):
  netgrp = parser.add_argument_group( 'Network' )
  netgrp.add_argument( '--bsmserver',
            dest='bsmserver', action='store',
            type=str, default='',
            help='Bokeh service manager connection details')
  netgrp.add_argument( '--bokehid',
            dest='bokehid', action='store',
            type=int, default=-1,
            help='Bokeh server id')
  netgrp.add_argument( '--ppid',
            dest='ppid', action='store',
            type=int, default=-1,
            help='PID of the parent process' )
  bokehgrp = parser.add_argument_group( 'Bokeh' )
  bokehgrp.add_argument( '--log-file',
            dest='bokehlogfnm', action='store',
            type=str, default='',
            help='Bokeh log-file name')
  bokehgrp.add_argument( '--address',
            dest='address', action='store',
            type=str, default='',
            help='Bokeh server address')
  bokehgrp.add_argument( '--port',
            dest='port', action='store',
            type=int, default=-1,
            help='Bokeh server port')
  return parser

def _getDocUrl(server, app_path):
  address_string = 'localhost'
  if server.address is not None and server.address != '':
    address_string = server.address
  url = "http://%s:%d%s%s" % (address_string, server.port, server.prefix, app_path)
  return url

def StartBokehServer(application, args, attempts=20):
  basicConfig(filename=args['bokehlogfnm'])
  address = args['address']
  port = args['port']
  while attempts:
    attempts -= 1
    try:
      authstr = f"{address}:{port}"
      server = Server(application,address=address,
                      port=port,
                      allow_websocket_origin=[authstr,authstr.lower()])
      server.start()
      
      msg = dgbservmgr.Message()
      msg.sendObjectToAddress(args['bsmserver'],
                      'bokeh_started', {'bokehid': args['bokehid'],
                                        'bokehurl': _getDocUrl(server,"/"),
                                        'bokehpid': os.getpid()})

      server.io_loop.start()
      return
    except OSError as ex:
      if "Address already in use" in str(ex):
        port += 1
      else:
        raise ex
      
  raise Exception("Failed to find available port")
    




#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : June 2019
#
# Deep learning apply server
#
#

import sys
import argparse

parser = argparse.ArgumentParser(
          description='Server application of a trained machine learning model')
parser.add_argument( '-v', '--version',
            action='version',version='%(prog)s 2.0')
parser.add_argument( 'modelfile', type=argparse.FileType('r'),
                     help='The input trained model file' )
parser.add_argument( '--ppid',
                     dest='parentpid', action='store',
                     type=int, default=-1,
                     help='PID of the parent process' )
netgrp = parser.add_argument_group( 'Network' )
netgrp.add_argument( '--address',
            dest='addr', metavar='ADDRESS', action='store',
            type=str, default='localhost',
            help='Address to listen on' )
netgrp.add_argument( '--port',
            dest='port', action='store',
            type=int, default=65432,
            help='Port to listen on')
datagrp = parser.add_argument_group( 'Data' )
datagrp.add_argument( '--mldir',
            dest='mldir', nargs=1,
            help='Machine Learning Directory' )
loggrp = parser.add_argument_group( 'Logging' )
loggrp.add_argument( '--log',
            dest='logfile', metavar='file', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='Progress report output' )
loggrp.add_argument( '--syslog',
            dest='sysout', metavar='stdout', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='System log' )
# optional
parser.add_argument( '--fakeapply', dest='fakeapply', action='store_true',
                     default=False,
                     help="applies a numpy average instead of the model" )
parser.add_argument( '--local', dest='localserv', action='store_true',
                     default=False,
                     help="use a local network socket connection" )

args = vars(parser.parse_args())
from odpy.common import *
initLogging( args )
redirect_stdout()

# Start listening as quickly as possible
import selectors
import socket
sel = selectors.DefaultSelector()
local = args['localserv']
host,port = args['addr'], args['port']
if local:
  addr = str(port)
  try:
    os.unlink(addr)
  except OSError:
    if os.path.exists(addr):
      raise
  host = 'LOCAL'
  sockfam = socket.AF_UNIX
else:
  addr = (host, port)
  sockfam = socket.AF_INET
lsock = socket.socket(sockfam, socket.SOCK_STREAM)
# Avoid bind() exception: OSError: [Errno 48] Address already in use
#lsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
try:
  lsock.bind(addr)
  lsock.listen()
except Exception as e:
  log_msg( 'Connection error for port', port, 'on host', host )
  log_msg( e )
  raise e
std_msg("listening on", addr)
lsock.setblocking(True)
sel.register(lsock, selectors.EVENT_READ, data=None)


# Keep all lengthy operations below
import numpy as np
import os
import psutil
import signal
import threading
import traceback

from odpy.common import Timer
import dgbpy.deeplearning_apply_serverlib as applylib


def signal_handler(signal, frame):
  raise applylib.ExitCommand()
signal.signal(signal.SIGINT,signal_handler)

def timerCB():
  if not parentproc.is_running():
    os.kill( psutil.Process().pid, signal.SIGINT )

def accept_wrapper(sock,applier):
  conn, addr = sock.accept()  # Should be ready to read
  conn.setblocking(True)
  message = applylib.Message(sel, conn, addr, applier)
  sel.register(conn, selectors.EVENT_READ, data=message)

timer = Timer(15, timerCB)
parentproc = None
if 'parentpid' in args:
  ppid = args['parentpid']
  if ppid > 0:
    parentproc = psutil.Process( ppid )
    timer.start()

applier = None
try:
  if applier == None:
    applier = applylib.ModelApplier( args['modelfile'].name, args['fakeapply'] )
  lastmessage = False
  cont = True
  while cont:
    events = sel.select(timeout=300)
    for key, mask in events:
      if key.data is None:
        accept_wrapper(key.fileobj,applier)
      else:
        message = key.data
        try:
          message.process_events(mask)
        except Exception:
          log_msg( "main: error: exception for",
                  f"{message.addr}:\n{traceback.format_exc()}",
          )
          message.close()
        lastmessage = lastmessage or message.lastmessage
        cont = not lastmessage or len(events) > 1
        cont = cont
        if parentproc != None and not parentproc.is_running():
          cont = False
        if cont:
          applier = message.applier
except KeyboardInterrupt:
  std_msg('caught keyboard interrupt, exiting')
except applylib.ExitCommand:
  std_msg('Found dead parent, exiting')
finally:
  timer.cancel()
  sel.close()
